# frozen_string_literal: true

module Core
  module Extension
    # [public] Defines behavior that can be applied to objects. Behavior will only be applied to each object once by default.
    #
    # force - If true, behavior will be applied even if it has already been applied to the including object.
    #
    class Behavior
      def initialize(force: false, &block)
        @force = force
        @module = build_module(&block)
      end

      # [public] Applies this behavior to an object via extend.
      #
      def apply_extend(object)
        if force? || unapplied?(object)
          object.extend(@module)
        end
      end

      # [public] Applies this behavior to an object via include.
      #
      def apply_include(object)
        if force? || unapplied?(object)
          object.include(@module)
        end
      end

      private def force?
        @force == true
      end

      private def unapplied?(object)
        !applied?(object)
      end

      private def applied?(object)
        object.ancestors.include?(@module)
      end

      private def build_module(&block)
        Module.new {
          define_singleton_method(:extended) do |base|
            arguments = {
              extended: true,
              included: false
            }

            base.class_exec(**Behavior.conditional_keyword_arguments(arguments, block), &block)
          end

          define_singleton_method(:included) do |base|
            arguments = {
              extended: false,
              included: true
            }

            base.class_exec(**Behavior.conditional_keyword_arguments(arguments, block), &block)
          end
        }
      end

      class << self
        def conditional_keyword_arguments(arguments, callable)
          arguments.keep_if { |key, _value|
            callable.parameters.any? { |type, name|
              (type == :key || type == :keyreq || type == :keyrest) && name == key
            }
          }
        end
      end
    end
  end
end
