# frozen_string_literal: true

module Core
  module Extension
    # [public] Defines a dependency that can be applied to objects.
    #
    # flags - Changes how the dependencies are applied. Possible values include:
    #
    #     * `:definition` - Extends the definition of the including object.
    #
    #     * `:implementation` - Extends the implementation of the including object.
    #
    #   Dependencies are applied with the `:implementation` flag by default.
    #
    # dependency - The dependency to be applied to objects, following the rules defined by flags.
    #
    # prepend - If `true`, methods will be prepended.
    #
    class Dependency
      ALLOWED_FLAGS = %i[definition implementation].freeze
      ALLOWED_FLAGS_STRING = ALLOWED_FLAGS.map { |allowed_flag|
        "`#{allowed_flag.inspect}'"
      }.join(", ").freeze

      def initialize(*flags, dependency:, prepend: false)
        flags = flags.map(&:to_sym)
        enforce_allowed_flags(flags)

        @flags = flags
        @dependency = dependency
        @prepend = prepend
        @definition = @flags.include?(:definition)
        @implementation = @flags.include?(:implementation) || (!definition? && !prepend?)
      end

      # [public] Apply the defined dependencies to an object via extend.
      #
      def apply_extend(object)
        if prepend?
          object.singleton_class.prepend(@dependency) if implementation? || definition?
        elsif definition? || implementation?
          object.extend(@dependency)
        end
      end

      # [public] Apply the defined dependencies to an object via include.
      #
      def apply_include(object)
        if prepend?
          object.prepend(@dependency) if implementation?
          object.singleton_class.prepend(@dependency) if definition?
        else
          object.include(@dependency) if implementation?
          object.extend(@dependency) if definition?
        end
      end

      private def implementation?
        @implementation == true
      end

      private def definition?
        @definition == true
      end

      private def prepend?
        @prepend == true
      end

      private def enforce_allowed_flags(flags)
        flags.each do |flag|
          unless allowed_flag?(flag)
            raise ArgumentError, "Expected flag `#{flag.inspect}' to be one of: #{ALLOWED_FLAGS_STRING}"
          end
        end
      end

      private def allowed_flag?(flag)
        ALLOWED_FLAGS.include?(flag)
      end
    end
  end
end
