# frozen_string_literal: true

require_relative "../core/extension/behavior"
require_relative "../core/extension/dependency"

module Is
  # [public] Turn a module into a mixin with superpowers.
  #
  module Extension
    # [public] Restrict the extension to one or more object types.
    #
    def restrict(*types)
      allowed_types.concat(types).uniq!
    end

    # [public] Define a module to be extended or included into objects that include this extension.
    #
    # TODO: These flags almost need different names, since instance extensions could technically end up on a class.
    # Name them by their intended use? Like a definition and usage lifecycle or something.
    #
    # Maybe `definition` and `implementation`?
    #
    def extends(*flags, dependencies: [], prepend: false, &block)
      dependencies.each do |dependency|
        defined_dependencies << Core::Extension::Dependency.new(*flags, dependency: dependency, prepend: prepend)
      end

      if block
        defined_dependencies << Core::Extension::Dependency.new(*flags, dependency: Module.new(&block), prepend: prepend)
      end
    end

    # [public] Define behavior to be evaled on objects that include this extension.
    #
    def applies(force: false, &block)
      defined_behaviors << Core::Extension::Behavior.new(force: force, &block)
    end

    def extended(base)
      enforce_allowed_types(base)
      extend_defined_dependencies(base)
      extend_defined_behaviors(base)

      super
    end

    def included(base)
      enforce_allowed_types(base)
      include_defined_dependencies(base)
      include_defined_behaviors(base)

      super
    end

    private def enforce_allowed_types(object)
      return if allowed?(object)

      raise TypeError, "Expected `#{object}' to be a decendent of an allowed type: #{allowed_types_string}"
    end

    private def allowed_types_string
      allowed_types.map { |allowed_type|
        "`#{allowed_type.inspect}'"
      }.join(", ")
    end

    private def allowed?(object)
      allowed_types.empty? || allowed_types.any? { |allowed_type|
        object <= allowed_type
      }
    end

    private def extend_defined_behaviors(object)
      defined_behaviors.each do |behavior|
        behavior.apply_extend(object)
      end
    end

    private def extend_defined_dependencies(object)
      defined_dependencies.each do |dependency|
        dependency.apply_extend(object)
      end
    end

    private def include_defined_behaviors(object)
      defined_behaviors.each do |behavior|
        behavior.apply_include(object)
      end
    end

    private def include_defined_dependencies(object)
      defined_dependencies.each do |dependency|
        dependency.apply_include(object)
      end
    end

    private def allowed_types
      @__allowed_types ||= []
    end

    private def defined_behaviors
      @__defined_behaviors ||= []
    end

    private def defined_dependencies
      @__defined_dependencies ||= []
    end
  end
end
